@@ -19,11 +19,11 @@ impl Array {
             result
         } else {
             let backward_op: BackwardOp =
-                Arc::new(|c, x| vec![Some(&(-&c[0].reciprocal().powf(2.0)) * x)]);
+                Arc::new(|c, _, x| vec![Some(&(-&c[0].reciprocal().powf(2.0)) * x)]);
 
             result
                 .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
+                .with_backward_op(backward_op)
         }
     }
 
@@ -40,11 +40,11 @@ impl Array {
         if !self.is_tracked {
             result
         } else {
-            let backward_op: BackwardOp = Arc::new(move |c, x| vec![Some(&(&c[0] * 2.0) * x)]);
+            let backward_op: BackwardOp = Arc::new(move |c, _, x| vec![Some(&(&c[0] * 2.0) * x)]);
 
             result
                 .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
+                .with_backward_op(backward_op)
         }
     }
 
@@ -56,11 +56,11 @@ impl Array {
         if !self.is_tracked {
             result
         } else {
-            let backward_op: BackwardOp = Arc::new(|c, x| vec![Some(x * &c[0].reciprocal())]);
+            let backward_op: BackwardOp = Arc::new(|c, _, x| vec![Some(x * &c[0].reciprocal())]);
 
             result
                 .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
+                .with_backward_op(backward_op)
         }
     }
 
@@ -74,7 +74,7 @@ impl Array {
         if !self.is_tracked {
             result
         } else {
-            let backward_op: BackwardOp = Arc::new(move |c, x| {
+            let backward_op: BackwardOp = Arc::new(move |c, _, x| {
                 vec![Some(Array::from((
                     Arc::clone(&c[0].dimensions),
                     Arc::new(mul_values(&x.values, &cached)),
@@ -83,13 +83,13 @@ impl Array {
 
             result
                 .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
+                .with_backward_op(backward_op)
         }
     }
 
-    /// Sums along the last `skip_size` dimensions.
-    pub fn sum(&self, skip_size: usize) -> Array {
-        if skip_size == 0 {
+    /// Sums along the last `skip_count` dimensions.
+    pub fn sum(&self, skip_count: usize) -> Array {
+        if skip_count == 0 {
             return self.clone();
         }
 
@@ -97,54 +97,48 @@ impl Array {
             output_slice[0] = arrays[0].iter().sum();
         });
 
-        let leading_count = self.dimensions.len().saturating_sub(skip_size);
+        let leading_count = self.dimensions.len().saturating_sub(skip_count);
         let target_dimensions: Vec<usize> = self
             .dimensions
             .iter()
             .copied()
             .take(leading_count)
-            .chain(vec![1; skip_size])
+            .chain(vec![1; skip_count])
             .collect();
+        let target_clone = target_dimensions.clone();
 
-        let output_values = Array::sliced_op(
-            vec![self],
-            &op,
-            &self.dimensions,
-            &target_dimensions,
-            skip_size,
-        );
-
-        let output_dimensions = target_dimensions[0..leading_count + 1].to_vec();
-        let result = Array::from((output_dimensions, output_values));
-
-        if !self.is_tracked {
-            result
+        let backward_op: Option<BackwardOp> = if !self.is_tracked {
+            None
         } else {
-            let backward_op: BackwardOp = Arc::new(move |c, x| {
+            Some(Arc::new(move |c, _, x| {
                 let op: SlicedOp = Box::new(move |output_slice, arrays| {
+                    // propagate delta to each summed dimension
                     for output in output_slice.iter_mut() {
                         *output = arrays[0][0];
                     }
                 });
 
-                let output_values = Array::sliced_op(
+                vec![Some(Array::sliced_op(
                     vec![&x],
                     &op,
-                    &target_dimensions,
+                    None,
+                    &target_clone,
                     &c[0].dimensions,
-                    skip_size,
-                );
+                    skip_count,
+                    0,
+                ))]
+            }))
+        };
 
-                vec![Some(Array::from((
-                    Arc::clone(&c[0].dimensions),
-                    Arc::new(output_values),
-                )))]
-            });
-
-            result
-                .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
-        }
+        Array::sliced_op(
+            vec![self],
+            &op,
+            backward_op,
+            &self.dimensions,
+            &target_dimensions,
+            skip_count,
+            skip_count,
+        )
     }
 
     /// Sums all the values of the array.
@@ -166,51 +160,32 @@ impl<'a, 'b> ops::Add<&'b Array> for &'a Array {
             }
         });
 
-        let output_values = Array::sliced_op(vec![self, other], &op, &dimensions, &dimensions, 0);
-        let result = Array::from((Arc::new(dimensions), Arc::new(output_values)));
-
-        if !self.is_tracked && !other.is_tracked {
-            result
+        let backward_op: Option<BackwardOp> = if !self.is_tracked && !other.is_tracked {
+            None
         } else {
-            let backward_op: BackwardOp = if self.is_tracked && other.is_tracked {
-                Arc::new(move |_, x| {
-                    vec![
+            Some(Arc::new(move |_, t, x| {
+                vec![
+                    if t[0] {
                         Some(Array::from((
                             Arc::clone(&x.dimensions),
                             Arc::clone(&x.values),
-                        ))),
+                        )))
+                    } else {
+                        None
+                    },
+                    if t[1] {
                         Some(Array::from((
                             Arc::clone(&x.dimensions),
                             Arc::clone(&x.values),
-                        ))),
-                    ]
-                })
-            } else if self.is_tracked {
-                Arc::new(move |_, x| {
-                    vec![
-                        Some(Array::from((
-                            Arc::clone(&x.dimensions),
-                            Arc::clone(&x.values),
-                        ))),
-                        None,
-                    ]
-                })
-            } else {
-                Arc::new(move |_, x| {
-                    vec![
-                        None,
-                        Some(Array::from((
-                            Arc::clone(&x.dimensions),
-                            Arc::clone(&x.values),
-                        ))),
-                    ]
-                })
-            };
+                        )))
+                    } else {
+                        None
+                    },
+                ]
+            }))
+        };
 
-            result
-                .with_children(vec![self.clone(), other.clone()])
-                .with_backward_op(Some(backward_op))
-        }
+        Array::sliced_op(vec![self, other], &op, backward_op, &dimensions, &dimensions, 0, 0)
     }
 }
 
@@ -236,10 +211,10 @@ impl<'a> ops::Neg for &'a Array {
         if !self.is_tracked {
             result
         } else {
-            let backward_op: BackwardOp = Arc::new(move |_, x| vec![Some(-x)]);
+            let backward_op: BackwardOp = Arc::new(move |_, _, x| vec![Some(-x)]);
             result
                 .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
+                .with_backward_op(backward_op)
         }
     }
 }
@@ -266,10 +241,10 @@ impl<'a> ops::Mul<Float> for &'a Array {
         if !self.is_tracked {
             result
         } else {
-            let backward_op: BackwardOp = Arc::new(move |_, x| vec![Some(x * other)]);
+            let backward_op: BackwardOp = Arc::new(move |_, _, x| vec![Some(x * other)]);
             result
                 .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
+                .with_backward_op(backward_op)
         }
     }
 }
@@ -286,25 +261,19 @@ impl<'a, 'b> ops::Mul<&'b Array> for &'a Array {
             }
         });
 
-        let dimensions = Arc::new(dimensions);
-        let output_values = Array::sliced_op(vec![self, other], &op, &dimensions, &dimensions, 0);
-        let result = Array::from((Arc::clone(&dimensions), Arc::new(output_values)));
-
-        if !self.is_tracked && !other.is_tracked {
-            result
+        let backward_op: Option<BackwardOp> = if !self.is_tracked && !other.is_tracked {
+            None
         } else {
-            let backward_op: BackwardOp = if self.is_tracked && other.is_tracked {
-                Arc::new(|c, x| vec![Some(&c[1] * x), Some(&c[0] * x)])
-            } else if self.is_tracked {
-                Arc::new(|c, x| vec![Some(&c[1] * x), None])
-            } else {
-                Arc::new(|c, x| vec![None, Some(&c[0] * x)])
-            };
+            Some(Arc::new(move |c, t, x| {
+                vec![
+                    if t[0] { Some(&c[1] * x) } else { None },
+                    if t[1] { Some(&c[0] * x) } else { None },
+                ]
+            }))
+        };
 
-            result
-                .with_children(vec![self.clone(), other.clone()])
-                .with_backward_op(Some(backward_op))
-        }
+        let dimensions = Arc::new(dimensions);
+        Array::sliced_op(vec![self, other], &op, backward_op, &dimensions, &dimensions, 0, 0)
     }
 }
 
@@ -443,7 +412,7 @@ mod tests {
         assert_eq!(a.sum(2), arr![arr![21.0], arr![42.0]]);
 
         let mut result = 2.0 * &a.sum(2);
-                result.backward(None);
+        result.backward(None);
 
         let gradient_expect = arr![
             arr![arr![2.0, 2.0, 2.0], arr![2.0, 2.0, 2.0]],
