@@ -28,6 +28,7 @@
 
 mod arithmetic;
 mod linalg;
+mod nonlinearity;
 
 use crate::numbers::*;
 
@@ -183,7 +184,8 @@ type SlicedOp = Box<dyn Fn(&mut [Float], Vec<&[Float]>)>;
 /// The forward operation computes an operation with respect to inputs.
 pub type ForwardOp = Arc<dyn Fn(&[&Array]) -> Array + Send + Sync>;
 /// The backward operation computes deltas with respect to inputs.
-pub type BackwardOp = Arc<dyn Fn(&mut Vec<Array>, &Array) -> Vec<Option<Array>> + Send + Sync>;
+pub type BackwardOp =
+    Arc<dyn Fn(&mut Vec<Array>, &[bool], &Array) -> Vec<Option<Array>> + Send + Sync>;
 
 /// An n-dimensional differentiable array. Stored in row-major order.
 ///
@@ -389,8 +391,8 @@ impl Array {
     }
 
     /// Sets the backward operation of the array for the backward pass.
-    fn with_backward_op(mut self, backward_op: Option<BackwardOp>) -> Array {
-        self.backward_op = backward_op;
+    fn with_backward_op(mut self, backward_op: BackwardOp) -> Array {
+        self.backward_op = Some(backward_op);
         self
     }
 
@@ -429,6 +431,7 @@ impl Array {
         self.backward(Some(delta));
     }
 
+    /// Propagates the number of consumers to each array in the graph.
     fn propagate_consumers(&mut self) {
         for child in &mut *self.children.lock().unwrap() {
             if child.is_tracked {
@@ -462,15 +465,23 @@ impl Array {
             Some(x) => {
                 let mut children_guard = self.children.lock().unwrap();
 
-                for child in &mut *children_guard {
-                    child.stop_tracking();
-                }
+                // TODO some closure to simplify tracked, and untracked operations
+                let is_tracked: Vec<bool> = children_guard
+                    .iter_mut()
+                    .map(|c| {
+                        let is_tracked = c.is_tracked;
+                        c.stop_tracking();
+                        is_tracked
+                    })
+                    .collect();
 
-                let delta = (*x)(&mut children_guard, &mut delta);
+                let delta = (*x)(&mut children_guard, &is_tracked, &mut delta);
 
-                for child in &mut *children_guard {
-                    child.start_tracking();
-                }
+                children_guard
+                    .iter_mut()
+                    .zip(is_tracked)
+                    .filter(|(_, t)| *t)
+                    .for_each(|(c, _)| c.start_tracking());
 
                 let mut handles = Vec::new();
                 // start a new thread which will wait on all consumers
@@ -524,22 +535,33 @@ impl Array {
     /// last `skip_size` dimensions, and performing the operation on the slices, which must have
     /// length of the product of the skipped dimensions.
     ///
+    /// # Arguments
+    ///
+    /// * `arrays` - The arrays to perform the operations on.
+    /// * `op` - The `SlicedOp`, which takes in slices of the arrays, and modifies the output slice.
+    /// * `input_dimensions` - The target dimensions to broadcast inputs to.
+    /// * `output_dimensions` - The output dimensions of the operation.
+    /// * `skip_count` - The number of dimensions which form the input slices.
+    /// * `flatten_count` - The number of dimensions which are flattened in the output.
+    ///
     /// # Panics
     ///
     /// Panics if unable to broadcast the arrays to `input_dimensions`.
     fn sliced_op(
         arrays: Vec<&Array>,
         op: &SlicedOp,
+        backward_op: Option<BackwardOp>,
         input_dimensions: &[usize],
         output_dimensions: &[usize],
-        skip_size: usize,
-    ) -> Vec<Float> {
+        skip_count: usize,
+        flatten_count: usize,
+    ) -> Array {
         let is_dimensions_valid = arrays.iter().all(|a| {
             a.dimensions
                 .iter()
                 .rev()
-                .skip(skip_size)
-                .zip(input_dimensions.iter().rev().skip(skip_size))
+                .skip(skip_count)
+                .zip(input_dimensions.iter().rev().skip(skip_count))
                 .all(|(x, y)| *x == 1 || *x == *y)
         });
 
@@ -548,45 +570,39 @@ impl Array {
         }
 
         // total length of the leading values
-        let leading_length = input_dimensions.iter().rev().skip(skip_size).product();
+        let leading_length = input_dimensions.iter().rev().skip(skip_count).product();
         // total length of the output
         let output_length = output_dimensions.iter().product();
 
         // count of leading dimensions
-        let leading_count = input_dimensions.len().saturating_sub(skip_size);
-        // total length of the slicesslice
+        let leading_count = input_dimensions.len().saturating_sub(skip_count);
+        // total length of the slices
         let group_lengths: Vec<usize> = arrays
             .iter()
-            .map(|a| a.dimensions.iter().rev().take(skip_size).product())
+            .map(|a| a.dimensions.iter().rev().take(skip_count).product())
             .collect();
         let output_group_length: usize = output_dimensions.iter().skip(leading_count).product();
 
-        let mut indices = vec![0; leading_count];
+        let mut indices = vec![0; input_dimensions.len()];
         let mut output_values = vec![0.0; output_length];
         for _ in 0..leading_length {
-            // add zero indices to the skipped dimensions
-            let chained = indices
-                .iter()
-                .copied()
-                .chain(vec![0; input_dimensions.len() - indices.len()])
-                .collect::<Vec<usize>>();
             let slices = arrays
                 .iter()
                 .enumerate()
                 .map(|(i, a)| {
-                    let offset = flatten_indices(&chained, &a.dimensions);
+                    let offset = flatten_indices(&indices, &a.dimensions);
                     &a.values[offset..offset + group_lengths[i]]
                 })
                 .collect();
 
-            let output_offset = flatten_indices(&chained, &output_dimensions);
+            let output_offset = flatten_indices(&indices, &output_dimensions);
             let output_slice =
                 &mut output_values[output_offset..output_offset + output_group_length];
 
             op(output_slice, slices);
 
-            for j in 0..indices.len() {
-                let current = indices.len() - j - 1;
+            for j in 0..leading_count {
+                let current = leading_count - j - 1;
                 if indices[current] == input_dimensions[current] - 1 {
                     indices[current] = 0;
                 } else {
@@ -596,23 +612,33 @@ impl Array {
             }
         }
 
-        output_values
+        let mut output_dimensions = output_dimensions.to_vec();
+        if flatten_count > 0 {
+            let flattened = output_dimensions.iter().rev().take(flatten_count).product::<usize>();
+            output_dimensions = output_dimensions.iter().take(output_dimensions.len() - flatten_count).copied().chain(vec![flattened]).collect::<Vec<usize>>();
+        }
+        
+        let result = Array::from((output_dimensions, output_values));
+        if let Some(backward_op) = backward_op {
+            result.with_children(arrays.iter().cloned().cloned().collect()).with_backward_op(backward_op)
+        } else {
+            result
+        }
     }
 
     /// Flattens the array by summing along dimensions to match the target dimensions.
     fn flatten_to(self, dimensions: Arc<Vec<usize>>) -> Array {
         if self.dimensions == dimensions {
-            return self;
+            self
+        } else {
+            let op: SlicedOp = Box::new(move |output_slice: &mut [Float], arrays: Vec<&[Float]>| {
+                for (i, output) in output_slice.iter_mut().enumerate() {
+                    *output += arrays[0][i];
+                }
+            });
+
+            Array::sliced_op(vec![&self], &op, None, &self.dimensions, &*dimensions, 0, 0)
         }
-
-        let op: SlicedOp = Box::new(move |output_slice: &mut [Float], arrays: Vec<&[Float]>| {
-            for (i, output) in output_slice.iter_mut().enumerate() {
-                *output += arrays[0][i];
-            }
-        });
-
-        let output_values = Array::sliced_op(vec![&self], &op, &self.dimensions, &*dimensions, 0);
-        Array::from((dimensions, Arc::new(output_values)))
     }
 
     /// Computes an operation on arrays.
@@ -620,7 +646,7 @@ impl Array {
     /// # Arguments
     ///
     /// * `arrays` - The arrays to perform the operations on.
-    /// * `op` - The `ForwardOp`, which takes in the arrays, and outputs another Array.
+    /// * `op` - The `ForwardOp`, which takes in the arrays, and outputs another array.
     /// * `backward_op` - The `BackwardOp`, which takes in the arrays, and the delta, and outputs a
     /// new delta, with respect to each input. It is recommended that any array operations here are
     /// untracked, unless interested in higher order derivatives.
@@ -634,22 +660,29 @@ impl Array {
     /// # use corgi::numbers::*;
     /// # use corgi::array::*;
     /// # fn main () {
-    /// let op: ForwardOp = Arc::new(|x: &[&Array]| {
+    /// let mul: ForwardOp = Arc::new(|x: &[&Array]| {
     ///     Array::from((x[0].dimensions(), x[0].values().iter().zip(x[1].values()).map(|(x, y)| x * y).collect::<Vec<Float>>()))
     /// });
     ///
-    /// let op_clone = Arc::clone(&op);
-    /// let backward_op: BackwardOp = Arc::new(move |c: &mut Vec<Array>, x: &Array| {
+    /// let mul_clone = Arc::clone(&mul);
+    /// let backward_op: BackwardOp = Arc::new(move |children: &mut Vec<Array>, is_tracked: &[bool], delta: &Array| {
     ///     vec![
-    ///         Some(Array::op(&[&c[1], x], Arc::clone(&op_clone), None)),
-    ///         Some(Array::op(&[&c[0], x], Arc::clone(&op_clone), None)),
+    ///         if is_tracked[0] {
+    ///             Some(Array::op(&[&children[1], delta], Arc::clone(&mul_clone), None))
+    ///         } else {
+    ///             None
+    ///         },
+    ///         if is_tracked[1] {
+    ///             Some(Array::op(&[&children[0], delta], Arc::clone(&mul_clone), None))
+    ///         } else {
+    ///             None
+    ///         }
     ///     ]
     /// });
     ///
-    /// let a = arr![1.0, 2.0, 3.0];
-    /// let b = arr![3.0, 2.0, 1.0];
-    ///
-    /// let mut product = Array::op(&vec![&a, &b], op, Some(backward_op));
+    /// let a = arr![1.0, 2.0, 3.0].tracked();
+    /// let b = arr![3.0, 2.0, 1.0].tracked();
+    /// let mut product = Array::op(&vec![&a, &b], mul, Some(backward_op));
     /// assert_eq!(product, arr![3.0, 4.0, 3.0]);
     /// product.backward(None);
     /// assert_eq!(product.gradient().unwrap(), arr![1.0, 1.0, 1.0]);
@@ -659,79 +692,14 @@ impl Array {
     /// ```
     pub fn op(arrays: &[&Array], op: ForwardOp, backward_op: Option<BackwardOp>) -> Array {
         let result = op(arrays);
-        result
-            .with_children(
-                arrays
-                    .iter()
-                    .map(|v| {
-                        let mut result = (*v).clone();
-                        result.is_tracked = v.is_tracked;
-                        result
-                    })
-                    .collect(),
-            )
-            .with_backward_op(backward_op)
-    }
-
-    /// Computes the ReLU of the array, defined as max(0, x) for all elements x in the array.
-    pub fn relu(&self) -> Array {
-        let (values, derivative) = self
-            .values
-            .iter()
-            .map(|&x| if x > 0.0 { (x, 1.0) } else { (0.0, 0.0) })
-            .unzip();
-
-        let result = Array::from((Arc::clone(&self.dimensions), Arc::new(values)));
-        let derivative = Array::from((Arc::clone(&self.dimensions), Arc::new(derivative)));
-
-        if !self.is_tracked {
+        if let Some(backward_op) = backward_op {
             result
+                .with_children(arrays.iter().map(|v| (*v).clone()).collect())
+                .with_backward_op(backward_op)
         } else {
-            let backward_op: BackwardOp = Arc::new(move |_, x| vec![Some(&derivative * x)]);
-
             result
-                .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
         }
     }
-
-    /// Computes the sigmoid operation on each value of the array.
-    pub fn sigmoid(&self) -> Array {
-        let values = Arc::new(
-            self.values
-                .iter()
-                .map(|x| 1.0 / (1.0 + (-x).exp()))
-                .collect::<Vec<Float>>(),
-        );
-
-        let cached = Arc::clone(&values);
-        let result = Array::from((Arc::clone(&self.dimensions), values));
-
-        if !self.is_tracked {
-            result
-        } else {
-            let backward_op: BackwardOp = Arc::new(move |c, x| {
-                let values = arithmetic::mul_values(
-                    &cached.iter().map(|v| v * (1.0 - v)).collect::<Vec<Float>>(),
-                    &x.values,
-                );
-                vec![Some(Array::from((
-                    Arc::clone(&c[0].dimensions),
-                    Arc::new(values),
-                )))]
-            });
-
-            result
-                .with_children(vec![self.clone()])
-                .with_backward_op(Some(backward_op))
-        }
-    }
-
-    /// Computes the softmax of the array.
-    pub fn softmax(&self) -> Array {
-        let exponentials = self.exp();
-        &exponentials / &exponentials.sum(1)
-    }
 }
 
 impl Clone for Array {
@@ -940,54 +908,6 @@ mod tests {
         assert_eq!(a.gradient().unwrap(), arr![2.0, 4.0, 6.0]);
     }
 
-    #[test]
-    fn test_softmax() {
-        let a = arr![
-            arr![(2.0 as Float).ln(), (2.0 as Float).ln()],
-            arr![(1.0 as Float).ln(), (1.0 as Float).ln()]
-        ]
-        .tracked();
-        let b = arr![arr![3.0, 5.0], arr![2.0, 5.0]].tracked();
-        let c = a.softmax().tracked();
-
-        let mut result = &c * &b;
-        assert_eq!(result, arr![arr![1.5, 2.5], arr![1.0, 2.5]]);
-
-        result.backward(None);
-        assert_eq!(c.gradient().unwrap(), arr![arr![3.0, 5.0], arr![2.0, 5.0]]);
-        assert_eq!(b.gradient().unwrap(), arr![arr![0.5, 0.5], arr![0.5, 0.5]]);
-        assert_eq!(
-            a.gradient().unwrap(),
-            arr![arr![-0.5, 0.5], arr![-0.75, 0.75]]
-        );
-    }
-
-    #[test]
-    fn test_relu() {
-        let a = arr![1.0, -2.0, 0.0].tracked();
-
-        let mut result = a.relu();
-        assert_eq!(result, arr![1.0, 0.0, 0.0]);
-
-        result.backward(None);
-        assert_eq!(a.gradient().unwrap(), arr![1.0, 0.0, 0.0]);
-    }
-
-    #[test]
-    fn test_sigmoid() {
-        let a = arr![arr![(3.0 as Float).ln()]].tracked();
-        let b = arr![arr![5.0]].tracked();
-        let c = a.sigmoid().tracked();
-
-        let mut result = &c * &b;
-        assert_relative_eq!(result, arr![arr![3.75]]);
-
-        result.backward(None);
-        assert_relative_eq!(c.gradient().unwrap(), arr![arr![5.0]]);
-        assert_relative_eq!(b.gradient().unwrap(), arr![arr![0.75]]);
-        assert_relative_eq!(a.gradient().unwrap(), arr![arr![0.9375]]);
-    }
-
     #[test]
     fn test_propagate() {
         let a = arr![5.0].tracked();
@@ -1026,8 +946,11 @@ mod tests {
         let b = arr![2.0].tracked();
 
         let product = &a * &b;
-        let result =
-            (*product.backward_op.unwrap())(&mut vec![a.clone(), b.clone()], &mut arr![1.0]);
+        let result = (*product.backward_op.unwrap())(
+            &mut vec![a.clone(), b.clone()],
+            &[true, true],
+            &mut arr![1.0],
+        );
 
         assert_eq!(result.len(), 2);
         assert_eq!(
@@ -1043,18 +966,6 @@ mod tests {
         assert!(product.is_tracked);
     }
 
-    #[test]
-    fn test_backward_untracked() {
-        let a = arr![5.0];
-        let b = arr![2.0].tracked();
-
-        let mut product = &a * &b;
-
-        product.backward(None);
-        assert_eq!(product.gradient().unwrap(), arr![1.0]);
-        assert_eq!(b.gradient().unwrap(), arr![5.0]);
-    }
-
     #[test]
     fn test_backward_untracked_both() {
         let a = arr![5.0];
@@ -1198,4 +1109,27 @@ mod tests {
         assert_eq!(b.gradient().unwrap(), arr![1.0, 1.0]);
         assert_eq!(a.gradient().unwrap(), arr![15.0, 30.0]);
     }
+
+    #[test]
+    fn test_backward_untracked() {
+        let a = arr![1.0, 2.0].tracked();
+        let b = arr![2.0, 1.0];
+        let mut product = &a * &b;
+
+        product.backward(None);
+        assert!(b.gradient().is_none());
+        assert!(a.gradient().is_some());
+    }
+
+    #[test]
+    fn test_backward_untracked_clone() {
+        let a = arr![1.0, 2.0].tracked();
+        let b = arr![2.0, 1.0];
+        let product = &a * &b;
+        let mut clone = product.clone();
+
+        clone.backward(None);
+        assert!(b.gradient().is_none());
+        assert!(a.gradient().is_some());
+    }
 }
